from ..ssp_utils import pegase_util as pegase_lib, \
    miles_util as miles_lib, \
    pegase_interp_util as pegase_interp_lib, \
    conroyinmiles_util as conroyinmiles_lib, \
    xshooter_util as xshooter_lib
from . import utils
import logging
import numpy as np
from scipy import ndimage
from multiprocessing import Pool



def process_Oversampling_templates(star, factor, i, j, k):
    starNew = ndimage.interpolation.zoom(star, factor, order=3)  # Oversampling
    return starNew, i, j, k


class model:

    def __init__(self, templateDir, instrumentDir, ssp_params, other_params):

        # setup some ssp_params
        self.ssp_name = ssp_params['model']
        self.imf = ssp_params['imf']
        self.slope = ssp_params['slope']
        self.isochrone = ssp_params['isochrone']
        self.single_alpha = ssp_params['single_alpha']
        self.factor = ssp_params['factor']
        self.FWHM_gal = ssp_params['FWHM_gal']
        self.FWHM_tem = ssp_params['FWHM_tem']
        self.dlam = ssp_params['dlam']
        self.age_range = ssp_params['age_range']
        self.metal_range = ssp_params['metal_range']
        self.wave_range = ssp_params['wave_range']
        self.interpolator_method = ssp_params['spec_interpolator']
        self.ncpu = other_params['ncpu']


        models = []

        if self.ssp_name=='miles':

            if self.single_alpha == False:
                models.append(miles_lib.miles(templateDir + 'miles/' + 'MILES_' + self.isochrone + '_' + self.imf + '_Ep0.00/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          age_range=self.age_range, metal_range=self.metal_range))
                models.append(miles_lib.miles(templateDir + 'miles/' + 'MILES_' + self.isochrone + '_' + self.imf + '_Ep0.40/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          age_range=self.age_range, metal_range=self.metal_range))
                alpha_grid = np.array([0, 0.4])

            elif self.single_alpha == 'alpha0':
                models.append(miles_lib.miles(templateDir + 'miles/' + 'MILES_' + self.isochrone + '_' + self.imf + '_Ep0.00/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

            else:
                models.append(miles_lib.miles(templateDir + 'miles/' + 'MILES_' + self.isochrone + '_' + self.imf + '_baseFe/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

        if self.ssp_name=='miles_interp':

            if self.single_alpha == False:
                models.append(miles_lib.miles(templateDir + 'miles_interp/' + 'MILES_' + self.isochrone + '_' + self.imf + '_Ep0.00/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          age_range=self.age_range, metal_range=self.metal_range))
                models.append(miles_lib.miles(templateDir + 'miles_interp/' + 'MILES_' + self.isochrone + '_' + self.imf + '_Ep0.40/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          age_range=self.age_range, metal_range=self.metal_range))
                alpha_grid = np.array([0, 0.4])

            elif self.single_alpha == 'alpha0':
                models.append(miles_lib.miles(templateDir + 'miles/' + 'MILES_' + self.isochrone + '_' + self.imf + '_Ep0.00/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

            else:
                models.append(miles_lib.miles(templateDir + 'miles_interp/' + 'MILES_' + self.isochrone + '_' + self.imf + '_baseFe/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

        if self.ssp_name=='miles_loginterp':

            if self.single_alpha == False:
                models.append(miles_lib.miles(templateDir + 'miles_loginterp/' + 'MILES_' + self.isochrone + '_' + self.imf + '_Ep0.00/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          age_range=self.age_range, metal_range=self.metal_range))
                models.append(miles_lib.miles(templateDir + 'miles_loginterp/' + 'MILES_' + self.isochrone + '_' + self.imf + '_Ep0.40/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          age_range=self.age_range, metal_range=self.metal_range))
                alpha_grid = np.array([0, 0.4])

            elif self.single_alpha == 'alpha0':
                models.append(miles_lib.miles(templateDir + 'miles/' + 'MILES_' + self.isochrone + '_' + self.imf + '_Ep0.00/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

            else:
                models.append(miles_lib.miles(templateDir + 'miles_loginterp/' + 'MILES_' + self.isochrone + '_' + self.imf + '_baseFe/M' + self.imf.lower() + '%.2f' % self.slope + '*.fits',
                                          FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))


        elif self.ssp_name == 'pegasehr_interp':

            assert self.single_alpha == True, "No alpha-enhanced model avaliable for PEGASE-HR yet."
            models.append(pegase_interp_lib.pegase_interp(templateDir + 'pegasehr/' + 'PEGASEHR_' + self.isochrone + '_' + self.imf + '_baseFe/',
                                FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

        elif self.ssp_name == 'pegasehr':

            assert self.single_alpha == True, "No alpha-enhanced model avaliable for PEGASE-HR yet."
            models.append(pegase_lib.pegase(templateDir + 'pegasehr/' + 'PEGASEHR_' + self.isochrone + '_' + self.imf + '_baseFe/',
                                            FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))


        elif self.ssp_name == 'conroyinmiles':

            if self.single_alpha == False:
                alpha_grid = np.load(templateDir + 'conroy/ssp_mist_v2.3_milesResUnit/' + 'miles/' + 'alpha_grid.npy')
                for alpha_i in range(len(alpha_grid)):
                    models.append(conroyinmiles_lib.conroy(templateDir + 'conroy/ssp_mist_v2.3_milesResUnit/' + 'miles/' + 'alpha_%s/' % int(alpha_i),
                                                           FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))
            else:
                # Just load alpha/fe=0.0 dex templates
                models.append(conroyinmiles_lib.conroy(templateDir + 'conroy/ssp_mist_v2.3_milesResUnit/' + 'miles/' + 'alpha_0/',
                                                       FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

        elif self.ssp_name == 'conroyinmiles_moremetalalpha':

            if self.single_alpha == False:
                alpha_grid = np.load(templateDir + 'conroy/ssp_mist_v2.3_milesResUnit_moremetalalpha/' + 'alpha_grid.npy')
                for alpha_i in range(len(alpha_grid)):
                    models.append(conroyinmiles_lib.conroy(templateDir + 'conroy/ssp_mist_v2.3_milesResUnit_moremetalalpha/' + 'alpha_%s/' % int(alpha_i),
                                                           FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))
            else:
                # Just load alpha/fe=0.0 dex templates
                models.append(conroyinmiles_lib.conroy(templateDir + 'conroy/ssp_mist_v2.3_milesResUnit_moremetalalpha/' + 'alpha_0/',
                                                       FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

        elif self.ssp_name == 'conroyinmilesCaT':

            if self.single_alpha == False:
                alpha_grid = np.load(templateDir + 'conroy/ssp_mist_v2.3_milesconroyCaTResUnit/' + 'alpha_grid.npy')
                for alpha_i in range(len(alpha_grid)):
                    models.append(conroyinmiles_lib.conroy(templateDir + 'conroy/ssp_mist_v2.3_milesconroyCaTResUnit/' + 'alpha_%s/' % int(alpha_i),
                                                           FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))
            else:
                # Just load alpha/fe=0.0 dex templates
                models.append(conroyinmiles_lib.conroy(templateDir + 'conroy/ssp_mist_v2.3_milesconroyCaTResUnit/' + 'alpha_0/',
                                                       FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

        elif self.ssp_name == 'conroyinmilesCaT_moremetalalpha':

            if self.single_alpha == False:
                alpha_grid = np.load(templateDir + 'conroy/ssp_mist_v2.3_milesconroyCaTResUnit_moremetalalpha/' + 'alpha_grid.npy')
                for alpha_i in range(len(alpha_grid)):
                    models.append(conroyinmiles_lib.conroy(templateDir + 'conroy/ssp_mist_v2.3_milesconroyCaTResUnit_moremetalalpha/' + 'alpha_%s/' % int(alpha_i),
                                                           FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))
            else:
                # Just load alpha/fe=0.0 dex templates
                models.append(conroyinmiles_lib.conroy(templateDir + 'conroy/ssp_mist_v2.3_milesconroyCaTResUnit_moremetalalpha/' + 'alpha_0/',
                                                       FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

        elif self.ssp_name == 'xshooter':
            assert self.single_alpha == True, "No alpha-enhanced model avaliable for X-shooter yet."
            models.append(xshooter_lib.xshooter(templateDir + 'xshooter/' + 'XSHOOTER_' + self.isochrone + '_' + self.imf + '_baseFe/XSL_SSP_*.fits',
                                          FWHM_tem=self.FWHM_tem, age_range=self.age_range, metal_range=self.metal_range))

        templates = np.stack([model.templates for model in models], axis=3)

        # Define some constants from SSP parameters
        reg_dim = templates.shape[1:]
        c = 299792.458 # speed of light in km/s
        wave = models[0].lam_temp
        velscale = c * np.log(wave[1]/wave[0])
        age_grid = models[0].age_grid[:, 0]
        metal_grid = models[0].metal_grid[0, :]
        age_grid_2d = models[0].age_grid
        metal_grid_2d = models[0].metal_grid

        self.FWHM_tem = models[0].FWHM_tem


        # This is for calculating the bins in age, metallicity
        xgrid = np.log10(age_grid) + 9
        ygrid = metal_grid
        x = xgrid
        y = ygrid
        xb = (x[1:] + x[:-1])/2  # internal grid borders
        yb = (y[1:] + y[:-1])/2
        xb = np.hstack([1.5*x[0] - x[1]/2, xb, 1.5*x[-1] - x[-2]/2])  # 1st/last border
        yb = np.hstack([1.5*y[0] - y[1]/2, yb, 1.5*y[-1] - y[-2]/2])


        if self.single_alpha == False:
            zgrid = alpha_grid
            z = zgrid
            zb = (z[1:] + z[:-1])/2
            zb = np.hstack([1.5*z[0] - z[1]/2, zb, 1.5*z[-1] - z[-2]/2])
        else:
            # This will be used when generating the interpolator,
            # if single_alpha==False, it will be still fine to get a 3D data
            alpha_grid = np.array([0])
            zb = np.array([0])



        # Setup FWHM_gal, FWHM_tem file and calculate sig
        if type(self.FWHM_gal) == str:
            logging.info('Load the line-spread function %s from the folder ./instrument' % self.FWHM_gal)
            lsf_gal = np.genfromtxt(instrumentDir + self.FWHM_gal)
        else:
            lsf_gal = np.zeros([len(wave), 2])
            lsf_gal[:, 0] = wave
            lsf_gal[:, 1] = self.FWHM_gal
        if type(self.FWHM_tem) == str:
            logging.info('Load the line-spread function %s from the folder ./instrument' % self.FWHM_tem)
            lsf_tem = np.genfromtxt(instrumentDir + self.FWHM_tem)
        else:
            lsf_tem = np.zeros([len(wave), 2])
            lsf_tem[:, 0] = wave
            lsf_tem[:, 1] = self.FWHM_tem
        sig = utils.cal_degrade_sig(np.interp(wave, lsf_gal[:, 0], lsf_gal[:, 1], left=np.nan, right=np.nan),
                                    np.interp(wave, lsf_tem[:, 0], lsf_tem[:, 1], left=np.nan, right=np.nan),
                                    np.diff(wave)[0])

        if np.all(np.isnan(sig)) == True:
            raise ValueError("The outputs FWHM_gal is lower than the SSP FWHM_tem.")
        elif np.any(np.isnan(sig)) == True:
            logging.info("FWHM_gal is smaller than the SSP FWHM_tem in some wavelength pixels, will remove them and contiue the program...")
            mask_nansig = ~np.isnan(sig)
            wave = wave[mask_nansig]
            templates = templates[mask_nansig, :, :]
            sig = sig[mask_nansig]
        else:
            logging.info("FWHM_gal is larger than the SSP FWHM_tem for all wavelength pixels, continue the program...")


        self.templates = templates
        self.reg_dim = reg_dim
        self.wave = wave
        self.velscale = velscale
        self.age_grid = age_grid
        self.logage_grid = np.log10(age_grid) + 9
        self.metal_grid = metal_grid
        self.alpha_grid = alpha_grid
        self.age_grid_2d = age_grid_2d
        self.metal_grid_2d = metal_grid_2d
        self.xb = xb
        self.yb = yb
        self.zb = zb
        self.lsf_gal = lsf_gal
        self.lsf_tem = lsf_tem
        self.sig = sig

        # Add the new wave
        if self.dlam == None:
            self.new_wave = self.wave
        else:
            self.new_wave = np.arange(wave[0], wave[-1], self.dlam)


        logging.info('==================SSP info==================')
        logging.info('model name:            %s' % self.ssp_name)
        logging.info('IMF:                   %s' % self.imf)
        logging.info('isochrone:             %s' % self.isochrone)
        logging.info('slope:                 %s' % self.slope)
        logging.info('single_alpha:          %s' % self.single_alpha)
        logging.info('factor:                %s' % self.factor)
        logging.info('FWHM_gal:              %s' % self.FWHM_gal)
        logging.info('FWHM_tem:              %s' % self.FWHM_tem)
        logging.info('dlam:                  %s' % self.dlam)
        logging.info('velscale (km/s):       %.2f' % velscale)
        logging.info('age_range (Gyr):       %s' % [float('{:.3f}'.format(i)) for i in age_grid[[0, -1]]])
        logging.info('metal_range (dex):     %s' % [float('{:.3f}'.format(i)) for i in metal_grid[[0, -1]]])
        logging.info('alpha_range (dex):     %s' % [float('{:.3f}'.format(i)) for i in alpha_grid[[0, -1]]])
        logging.info('templates dim:         %s' % list(reg_dim))
        logging.info('n_templates:           %s' % np.prod(reg_dim))
        logging.info('SSP wave range:        %s' % [float('{:.3f}'.format(i)) for i in [self.wave[0], self.wave[-1]]])
        if self.wave_range==None:
            logging.info('Cube wave range:       %s' % [float('{:.3f}'.format(i)) for i in [self.new_wave[0], self.new_wave[-1]]])
        else:
            logging.info('Cube wave range:       %s' % [float('{:.3f}'.format(i)) for i in self.wave_range])
        logging.info('============================================')






    def oversample(self):
        # Oversample the templates and integral the spectra again
        wave_oversampled = ndimage.interpolation.zoom(self.wave, self.factor, order=1)
        templatesOversampled = np.zeros([self.templates.shape[0]*self.factor] + list(self.templates.shape[1:]))


        # Setup FWHM_gal file and calculate sig
        sig_oversampled = utils.cal_degrade_sig(np.interp(wave_oversampled, self.lsf_gal[:, 0], self.lsf_gal[:, 1]),
                                                np.interp(wave_oversampled, self.lsf_tem[:, 0], self.lsf_tem[:, 1]),
                                                np.diff(wave_oversampled)[0])


        pool = Pool(processes=self.ncpu)
        results = []
        for i in range(templatesOversampled.shape[1]):
            for j in range(templatesOversampled.shape[2]):
                for k in range(templatesOversampled.shape[3]):
                    results.append(pool.apply_async(process_Oversampling_templates, (self.templates[:, i, j, k], self.factor, i, j, k, )))
        pool.close()
        pool.join()

        for result in results:
            starNew, i, j, k = result.get()
            templatesOversampled[:, i, j, k] = starNew

        self.templatesOversampled = templatesOversampled
        self.wave_oversampled = wave_oversampled
        self.sig_oversampled = sig_oversampled

        self.templatesTransposed = np.transpose(self.templatesOversampled, (1, 2, 3, 0))



import os
import time
import logging
import numpy as np
from multiprocessing import Pool
from time import perf_counter as clock

from scipy.stats import binned_statistic_2d

import matplotlib.colors as colors
from matplotlib import pyplot as plt


def setupLogfile(logfile, __version__, mode='a', welcome=True):
    """Initialise the LOGFILE."""
    welcomeString = "\n\n# ============================================== #\n#{:^48}#\n#{:^48}#\n# ============================================== #\n".format(
        "GalCraft", "Version " + __version__
    )

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    logging.basicConfig(
        handlers=[
            logging.FileHandler(logfile, mode=mode),
            logging.StreamHandler()
        ],
        level=logging.INFO,
        format="%(asctime)s - %(levelname)-8s - %(module)s: %(message)s",
        datefmt="%m/%d/%y %H:%M:%S",
    )
    logging.Formatter.converter = time.gmtime
    if welcome == True:
        logging.info(welcomeString)


##################################################################################
# Initialization functions
def initialize_Dirs(CommandOptions):
    '''
    Initialize directory paths
    :param CommandOptions: Command parameters
    :return: Paths for each directory
    '''
    if os.path.isfile(CommandOptions.defaultDir) == True:
        for line in open(CommandOptions.defaultDir, "r"):
            if not line.startswith('#'):
                line = line.split('=')
                line = [x.strip() for x in line]
                if os.path.isdir(line[1]) == True:
                    if line[0] == 'configDir':
                        configDir = line[1]
                    elif line[0] == 'outputDir':
                        outputDir = line[1]
                    elif line[0] == 'modelDir':
                        modelDir = line[1]
                    elif line[0] == 'templateDir':
                        templateDir = line[1]
                else:
                    print("WARNING! "+line[1]+" specified as default "+line[0]+" is not a directory!")
    else:
        print("WARNING! "+CommandOptions.defaultDir+" is not a file!")

    return configDir, modelDir, templateDir, outputDir


##################################################################################
# Doppler shift functions

def doppler_shift(wavelength, flux, dv):
    '''
    This is the function from The Payne
    dv is in km/s
    We use the convention where a positive dv means the object is moving away.

    This linear interpolation is actually not that accurate, but is fine if you
    only care about accuracy to the level of a few tenths of a km/s. If you care
    about better accuracy, you can do better with spline interpolation.
    '''
    c = 2.99792458e5 # km/s
    doppler_factor = np.sqrt((1 - dv/c)/(1 + dv/c))
    new_wavelength = wavelength * doppler_factor
    new_flux = np.interp(new_wavelength, wavelength, flux)
    return new_flux



##################################################################################
# Reddening function
def reddening_cal00(lam, ebv):
    """
    Reddening curve of `Calzetti et al. (2000)
    <http://ui.adsabs.harvard.edu/abs/2000ApJ...533..682C>`_
    This is reliable between 0.12 and 2.2 micrometres.
    - LAMBDA is the restframe wavelength in Angstrom of each pixel in the
      input galaxy spectrum (1 Angstrom = 1e-4 micrometres)
    - EBV is the assumed E(B-V) colour excess to redden the spectrum.
      In outputs the vector FRAC gives the fraction by which the flux at each
      wavelength has to be multiplied, to model the dust reddening effect.

    """
    ilam = 1e4/lam  # Convert Angstrom to micrometres and take 1/lambda
    rv = 4.05  # C+00 equation (5)

    # C+00 equation (3) but extrapolate for lam > 2.2
    # C+00 equation (4) (into Horner form) but extrapolate for lam < 0.12
    k1 = rv + np.where(lam >= 6300, 2.76536*ilam - 4.93776,
                       ilam*((0.029249*ilam - 0.526482)*ilam + 4.01243) - 5.7328)
    fact = 10**(-0.4*ebv*k1.clip(0))  # Calzetti+00 equation (2) with opposite sign

    return fact # The model spectrum has to be multiplied by this vector





##################################################################################
# Plotting functions

def plot_binned_grids_color(x, y, values, statistic, x_edges, y_edges, xlabel, ylabel,
                            plot_cb=True, cmap='hot', cblabel=None, color_Lognorm=False,
                            invert_x=True, **kwargs):
    '''

    :param x:
    :param y:
    :param values:
    :param statistic:
    :param x_edges:
    :param y_edges:
    :param cmap:
    :param xlabel:
    :param ylabel:
    :param cblabel:
    :param percentile:
    :param norm:
    :return:
    '''

    binned_statistic = binned_statistic_2d(x, y, values, statistic, bins=[x_edges, y_edges]).statistic

    if statistic == 'count':
        binned_statistic[binned_statistic==0] = np.nan

    vmin = np.nanpercentile(binned_statistic[binned_statistic!=0], 0.5)
    vmax = np.nanpercentile(binned_statistic[binned_statistic!=0], 99.5)

    if color_Lognorm == True:
        color_norm = colors.LogNorm(vmin, vmax)
    else:
        color_norm = colors.Normalize(vmin, vmax)

    ax = plt.gca()
    im = ax.pcolormesh(x_edges, y_edges, binned_statistic.T, cmap=cmap, norm=color_norm, **kwargs)
    if plot_cb == True:
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label(cblabel)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)
    if invert_x:
        ax.invert_xaxis()
    # fig.tight_layout()

    return im, ax, binned_statistic, x_edges, y_edges





def plot_parameter_maps(mass_fraction_pixel_bin, age_grid_2d, metal_grid_2d, alpha_grid, reg_dim,
                        x_edges, y_edges, xlabel, ylabel, cmap, **kwargs):
    '''

    :param mass_fraction_pixel_bin:
    :param age_grid_2d:
    :param metal_grid_2d:
    :param reg_dim:
    :param x_edges:
    :param y_edges:
    :param xlabel:
    :param ylabel:
    :param cmap:
    :param kwargs:
    :return:
    '''

    x_edges = np.flip(x_edges)

    mass_weighted_age = np.zeros(mass_fraction_pixel_bin.shape[3:])
    mass_weighted_metal = np.zeros(mass_fraction_pixel_bin.shape[3:])
    mass_weighted_alpha = np.zeros(mass_fraction_pixel_bin.shape[3:])

    for i in range(mass_fraction_pixel_bin.shape[-2]):
        for j in range(mass_fraction_pixel_bin.shape[-1]):
            mass_fraction_pixel = mass_fraction_pixel_bin[:, :, :, i, j]
            if np.sum(mass_fraction_pixel) == 0:
                mass_weighted_age[i, j] = np.nan
                mass_weighted_metal[i, j] = np.nan
                mass_weighted_alpha[i, j] = np.nan
            else:
                age, metal, alpha = mean_age_metal_alpha(age_grid_2d, metal_grid_2d, alpha_grid, mass_fraction_pixel, reg_dim, True)
                mass_weighted_age[i, j] = age
                mass_weighted_metal[i, j] = metal
                mass_weighted_alpha[i, j] = alpha

    plt.figure(figsize=[18, 4])

    plt.subplot(131)
    ax1 = plt.gca()
    vmin = np.nanpercentile(mass_weighted_age, 0.5)
    vmax = np.nanpercentile(mass_weighted_age, 99.5)
    im1 = ax1.pcolormesh(x_edges, y_edges, mass_weighted_age, cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)
    cbar1 = plt.colorbar(im1, ax=ax1)
    cbar1.set_label('log(Age)')
    ax1.set_ylabel(ylabel)
    ax1.set_xlabel(xlabel)

    plt.subplot(132)
    ax2 = plt.gca()
    vmin = np.nanpercentile(mass_weighted_metal, 0.5)
    vmax = np.nanpercentile(mass_weighted_metal, 99.5)
    im2 = ax2.pcolormesh(x_edges, y_edges, mass_weighted_metal, cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)
    cbar2 = plt.colorbar(im2, ax=ax2)
    cbar2.set_label('[M/H]')
    ax2.set_ylabel(ylabel)
    ax2.set_xlabel(xlabel)

    plt.subplot(133)
    ax3 = plt.gca()
    vmin = np.nanpercentile(mass_weighted_alpha, 0.5)
    vmax = np.nanpercentile(mass_weighted_alpha, 99.5)
    im3 = ax3.pcolormesh(x_edges, y_edges, mass_weighted_alpha, cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)
    cbar3 = plt.colorbar(im3, ax=ax3)
    cbar3.set_label('[Alpha/Fe]')
    ax3.set_ylabel(ylabel)
    ax3.set_xlabel(xlabel)

    ax1.invert_xaxis()
    ax2.invert_xaxis()
    ax3.invert_xaxis()

    plt.tight_layout()

##################################################################################





##################################################################################
# Spectral Degrading functions


def cal_degrade_sig(FWHM_gal, FWHM_tem, dlam):
    FWHM_dif = np.sqrt(FWHM_gal ** 2 - FWHM_tem ** 2)
    sigma = FWHM_dif / 2.355 / dlam

    # if np.all((sigma==0)):
    #     sigma = sigma - 1 # this is the case when FWHM_gal == FWHM_tem

    return sigma



def degrade_spec_ppxf(spec, spec_err=None, sig=0, gau_npix=None):
    '''
    Modified from PPXF v8.1.0
    :param spec:
    :param spec_err:
    :param sig:
    :param gau_npix:
    :return:
    '''
    # This function can now skip the err if doesn't have
    if np.all(sig == 0):
        return spec, spec_err
    elif np.isscalar(sig):
        sig = np.zeros(spec.shape) + sig
    sig = sig.clip(1e-10)  # forces zero sigmas to have 1e-10 pixels

    if gau_npix == None:
        p = int(np.ceil(np.max(3*sig)))
    else:
        p = gau_npix
    m = 2 * p + 1  # kernel sizes
    x2 = np.linspace(-p, p, m) ** 2

    n = spec.size
    a = np.zeros((m, n))
    for j in range(m):   # Loop over the small size of the kernel
        a[j, p:-p] = spec[j:n-m+j+1]

    gau = np.exp(-x2[:, None] / (2 * sig ** 2))
    gau /= np.sum(gau, 0)[None, :] # Normalize kernel

    conv_spectrum = np.einsum('ij,ij->j', a, gau)

    if np.all(spec_err != None):
        a_e = np.zeros((m, n))
        for j in range(m):
            a_e[j, p:-p] = spec_err[j:n - m + j + 1]
        conv_spectrum_err = np.sqrt((a_e**2 * gau**2).sum(0))
    else:
        conv_spectrum_err = None

    return conv_spectrum, conv_spectrum_err



def process_degrade_cube(spec, spec_err, i, j, sigma, gau_npix=None):
    spec_ij_degraded, spec_err_ij_degraded = degrade_spec_ppxf(spec, spec_err, sigma, gau_npix)
    return spec_ij_degraded, spec_err_ij_degraded, i, j



def degrade_spec_cube(cube_flux, cube_err, FWHM_gal, FWHM_tem, ncpu, dlam, gau_npix=None):
    # If not using cube_err, set it tobe np.zeros(cube_flux.shape)
    t = clock()

    sigma = cal_degrade_sig(FWHM_gal, FWHM_tem, dlam)

    cube_flux_degraded = np.zeros(cube_flux.shape)
    cube_err_degraded = np.zeros(cube_err.shape)

    pool = Pool(processes=ncpu)
    results = []
    for i in range(cube_flux.shape[1]):
        for j in range(cube_flux.shape[2]):
            results.append(pool.apply_async(process_degrade_cube, (cube_flux[:, i, j], cube_err[:, i, j], i, j, sigma, gau_npix)))
    pool.close()
    pool.join()

    for result in results:
        spec_ij_degraded, spec_err_ij_degraded, i, j = result.get()
        cube_flux_degraded[:, i, j] = spec_ij_degraded
        cube_err_degraded[:, i, j] = spec_err_ij_degraded

    print('Elapsed time in generating spectra: %.2f s' % (clock() - t))

    return cube_flux_degraded, cube_err_degraded





##################################################################################
# Other functions

def mean_age_metal_alpha(age_grid_2d, metal_grid_2d, alpha_grid, weights, reg_dim, quiet=False):
    '''
    This is modified from PPXF, compatible with single alpha case
    :param age_grid_2d:
    :param metal_grid_2d:
    :param alpha_grid:
    :param weights:
    :param reg_dim:
    :param quiet:
    :return:
    '''

    log_age_grids = np.log10(age_grid_2d) + 9
    metal_grids = metal_grid_2d

    weights_sum = np.sum(weights, axis=2)
    mean_log_age = np.sum(weights_sum * log_age_grids) / np.sum(weights_sum)
    mean_metal = np.sum(weights_sum * metal_grids) / np.sum(weights_sum)
    mean_alpha = np.sum(np.sum(weights * alpha_grid, axis=2)) / np.sum(weights_sum)

    if not quiet:
        print('Weighted <logAge> [yr]: %#.3g' % mean_log_age)
        print('Weighted <[M/H]>: %#.3g' % mean_metal)
        print('Weighted <[Alpha/Fe]>: %#.3g' % mean_alpha)

    return mean_log_age, mean_metal, mean_alpha




#!/usr/bin/env python

import os
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"
os.environ["OMP_NUM_THREADS"] = "1"
import warnings
warnings.filterwarnings("ignore")
import logging
import ebf
import json
import optparse
from time import perf_counter as clock
import multiprocessing

import GalCraft.modules.ssp_loader as ssp_loader
import GalCraft.modules.binner as binner
import GalCraft.modules.spec_generator as spec_generator
import GalCraft.modules.cube_maker as cube_maker
import GalCraft.modules.cot as cot
import GalCraft.modules.utils as utils
from GalCraft._version import __version__


def run_GalCraft(CommandOptions):



    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # - - - - - - - -   I N I T I A L I Z E   G A L C R A F T - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    t_init = clock()

    # - - - - - - - - - - INITIALIZATION - - - - - - - - - -

    # Initialize cube name and paths
    cube_name = CommandOptions.configName
    setup_cube_name = 'setup_' + cube_name
    configDir, modelDir, templateDir, outputDir = utils.initialize_Dirs(CommandOptions)
    # Load the Configuration file
    with open(configDir + setup_cube_name + '.json', 'r') as f:
        params = json.load(f)
    # Setup the outputs folder
    filepath = outputDir + cube_name + '/'
    if os.path.exists(filepath)==False: os.mkdir(filepath[:-1])
    # Setup logger
    if params['other_params']['mode'] == 'continue': # For continue mode, change the logger mode from 'w' to 'a'
        params['other_params']['log_mode'] = 'a'
    utils.setupLogfile(logfile=filepath + 'outputs.log', __version__=__version__, mode=params['other_params']['log_mode'], welcome=True)
    logging.info('Loaded the setup file (path %s)' % (configDir + setup_cube_name + '.json'))
    logging.info('Outputs DIR: %s' % filepath)
    # Setup cpu cores to use
    if params['other_params']['ncpu'] == None:
        # n_cores = len(os.sched_getaffinity(0))
        n_cores = multiprocessing.cpu_count()
        logging.info('Change the CPU cores from %s to %s due to the CPU availability.' % (params['other_params']['ncpu'], n_cores) )
        params['other_params']['ncpu'] = n_cores
    else:
        logging.info('Run the framework using %s CPU cores.' % params['other_params']['ncpu'])



    # - - - - - - - - - - INSTRUMENT SETUP - - - - - - - - - -

    # Obtain ./instrument/ dir
    this_dir, this_filename = os.path.split(__file__)
    instrumentDir = os.path.join(this_dir, "instrument") + '/'
    # Load instrument information
    # If "instrument"=="MUSE-WFM", it will load "MUSE-WFM.json" and use the "spatial_resolution" and "spatial_bin";
    # If "instrument"=="DIY", you will design your own instrument using "spatial_resolution" and "spatial_bin" in the configuration file;
    # If "instrument"=="DEFAULT", it will generate a huge cube using all the particles and only will be used in the configuration file.
    inst = params['cube_params']['instrument']
    if inst.upper() != 'DIY' and inst.upper()!='DEFAULT':
        with open(instrumentDir + inst + '.json', 'r') as f:
            inst_params = json.load(f)
        for key in inst_params:
            params['cube_params'][key] = inst_params[key]
        logging.info('INST mode, apply the %s instrument spatial properties from the preset file.' % inst)
    elif inst.upper() == 'DEFAULT':
        logging.info('DEFAULT mode, will use all the particles and only the given spatial resolution will be used.')
    else:
        logging.info("DIY mode, apply the given spatial resolution and bin number.")



    # - - - - - - - - - - LOAD SPECTRAL TEMPLATES - - - - - - - - - -

    logging.info('Loading SSP models...')
    t = clock()
    ssp_model = ssp_loader.model(templateDir, instrumentDir, params['ssp_params'], params['other_params'])
    ssp_model.oversample()
    logging.info('SSP model has been successfully loaded, time elapsed: %.2f s' % (clock() - t))



    # - - - - - - - - - - LOAD STELLAR CATALOG - - - - - - - - - -

    logging.info('Loading the E-Galaxia model from %s' % (modelDir + params['other_params']['model_name']))
    t = clock()
    d_t = ebf.read(modelDir + params['other_params']['model_name'],'/')
    # Check the need to locate/rotate the E-Galaxia model.
    if 'vr' in d_t.keys():
        logging.info('Model has already been rotated, no location/rotation transformation is applied.')
        for key in params['oparams']:
            params['oparams'][key] = None
    else:
        logging.info('Specifing the location/rotation of the galaxy')
        cot.observe(d_t, params['oparams'])
    # Check extinction values
    if params['cube_params']['use_extinc'] == True:
        assert 'exbv' in d_t.keys(), "No extinction in the loaded model, process terminated."
    # Check the need to apply the flux calibration due to the total Galaxy Stellar mass.
    if params['other_params']['nparticles'] != None:
        nparticles = params['other_params']['nparticles']
    else:
        if 'shuffle' in params['other_params']['model_name']:
            original_model_name = '_'.join([x for x in params['other_params']['model_name'][:-4].split('_') if 'shuffle' not in x]) + '.ebf'
            d_original = ebf.read(modelDir + original_model_name,'/')
            nparticles = len(d_original['px'])
            del d_original
        else:
            nparticles = len(d_t['vr'])
    # In case nparticles changed due to the "debug" mode.
    if params['other_params']['mode'] == 'debug':
        nparticles = 500000
    logging.info('Number of particles = %s or Galaxy Stellar Mass' % nparticles)
    # Calculate mass_cali_factor for flux calibration
    if params['other_params']['gal_mass'] != None:
        mass_cali_factor = params['other_params']['gal_mass'] / nparticles
        logging.info('Input Galaxy Stellar mass is %s, then the calibration factor is %s.' % (params['other_params']['gal_mass'], mass_cali_factor))
    else:
        logging.info('No flux calibration due to Galaxy Stellar mass is applied.')
        mass_cali_factor = 1
    logging.info('E-Galaxia model has been successfully loaded, time elapsed: %.2f s' % (clock() - t))




    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # - - - - - - - - - -   A N A L Y S I S   M O D U L E S   - - - - - - - - - - -
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



    # - - - - - - - - - - SPATIAL BINNING - - - - - - - - - -

    if params['other_params']['mode'] != 'continue':
        logging.info('Start binning the model...')
        t = clock()
        d_t_l, statistic_count_l, x_edges_l, y_edges_l = binner.spatial_binner(d_t, params['cube_params'], params['other_params'],
                                                                               ssp_model.age_grid, ssp_model.metal_grid, ssp_model.alpha_grid,
                                                                               filepath, configDir + setup_cube_name,
                                                                               params['oparams']['distance'], nparticles)
        logging.info('Binning process has been finished, time elapsed: %.2f s' % (clock() - t))
    else:
        logging.info('Continue the running from previous run...')
        t = clock()
        d_t_l, statistic_count_l, x_edges_l, y_edges_l = binner.spatial_binner_continue(params['cube_params'], params['other_params'],
                                                                                        filepath, configDir + setup_cube_name)
        logging.info('Binning process has been finished, time elapsed: %.2f s' % (clock() - t))
    d_t = None



    # - - - - - - - - - - START DATACUBE LOOP  - - - - - - - - - -

    for cube_idx in range(len(statistic_count_l)):
        d_t = d_t_l[cube_idx]
        statistic_count = statistic_count_l[cube_idx]
        x_edges = x_edges_l[cube_idx]
        y_edges = y_edges_l[cube_idx]



    # - - - - - - - - - - GENERATE DATACUBE  - - - - - - - - - -

        if params['other_params']['mode'] == 'continue' and os.path.exists(filepath + 'data_cube_' + str(cube_idx) + '.fits'):
            logging.info("Data Cube No.%s exists in %s, will not be run in 'continue' mode" % (cube_idx+1, filepath))
            continue
        else:
            logging.info('Start generating the datacube No.%s...' % (cube_idx+1))
            t = clock()
            Spec_DataCube = spec_generator.Spec_Generator(d_t, x_edges, y_edges, statistic_count, ssp_model,
                                                          params['cube_params'], params['ssp_params'], params['other_params'],
                                                          filepath, cube_idx)
            Spec_DataCube()
            data_cube = Spec_DataCube.data_cube
            data_cube = data_cube * mass_cali_factor
            logging.info('Calibrated the flux based on the input Galaxy Stellar Mass=%s.' % params['other_params']['gal_mass'])
            # Free memory
            Spec_DataCube = None
            d_t = None
            statistic_count = None
            logging.info('Stellar spectra have been generated, time elapsed: %.2f s' % (clock() - t))


    # - - - - - WRITE DATACUBE  - - - - -

        cube_maker.write_cube(data_cube, params, x_edges, y_edges, ssp_model.new_wave, filepath, cube_idx, ssp_model.velscale, __version__)
        logging.info('Data Cube No.%s has been written in %s, total time elapsed: %.2f s' % (cube_idx+1, filepath , clock() - t))




    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # - - - - - - - -  F I N A L I S E   T H E   A N A L Y S I S  - - - - - - - - -
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



    # - - - - - SAVE CONFIG  - - - - -

    logging.info('Write the json file into the outputs folder.')
    with open(filepath + 'setup.json', 'w') as f:
        json.dump(params, f, indent=2)
    logging.info('All the datacubes have been generated successfully, total time elapsed: %.2f s' % (clock() - t_init))
    logging.info("GalCraft completed successfully.")







# ============================================================================ #
#                           M A I N   F U N C T I O N                          #
# ============================================================================ #
def main(args=None):

    # Capture command-line arguments
    parser = optparse.OptionParser(usage="%GalCraft [options] arg")
    parser.add_option("--config",      dest="configName", type="string", \
            help="State the name of the MasterConfig file.")
    parser.add_option("--default-dir", dest="defaultDir", type="string", \
            help="File defining default directories for input, outputs, configuration files, and spectral templates.")
    (CommandOptions, args) = parser.parse_args()

    # Check if required command-line argument is given
    assert CommandOptions.configName != None, "Please specify the path of the Configuration.json file to be used. Exit!"

    # Run the framework
    run_GalCraft(CommandOptions)


if __name__ == '__main__':
    main()



import glob, re

import numpy as np
from scipy import interpolate
from astropy.io import fits


def age_metal(filename):
    """
    Extract age (T) and metallicity (Z or MH) values from the given filename.
    Supports formats:
    - 'XSL_SSP_logT<logT>_MH<MH>_Kroupa_PC.fits'
    - 'XSL_SSP_T<age>_Z<Z>_Kroupa_P00.fits'
    """
    pattern1 = r'XSL_SSP_logT(?P<logT>[-+]?[0-9]*\.?[0-9]+)_MH(?P<MH>[-+]?[0-9]*\.?[0-9]+)_.+\.fits'
    pattern2 = r'XSL_SSP_T(?P<T>[-+]?[0-9]*\.?[0-9]+e?[+-]?[0-9]*)_Z(?P<Z>[-+]?[0-9]*\.?[0-9]+)_.+\.fits'

    match1 = re.search(pattern1, filename)
    match2 = re.search(pattern2, filename)

    if match1:
        return 10 ** float(match1.group('logT')) / 1e9, float(match1.group('MH'))
    elif match2:
        return float(match2.group('T')) / 1e9, float(match2.group('Z'))
    else:
        return None

class xshooter:

    def __init__(self, pathname, FWHM_tem='lsf_xshooter', age_range=None,
                 metal_range=None):

        files = glob.glob(pathname)
        assert len(files) > 0, "Files not found %s" % pathname

        all = [age_metal(f) for f in files]
        all_ages, all_metals = np.array(all).T
        ages, metals = np.unique(all_ages), np.unique(all_metals)
        n_ages, n_metal = len(ages), len(metals)

        assert set(all) == set([(a, b) for a in ages for b in metals]), \
            'Ages and Metals do not form a Cartesian grid'

        if '_KU_' in files[0]:
            flux_factor = 5567946.09
        elif '_SA_' in files[0]:
            flux_factor = 9799552.50
        else:
            raise ValueError('initial-mass function is unknown, can not calculate flux unit factor')

        hdu = fits.open(files[0])
        ssp = hdu[0].data
        h2 = hdu[0].header
        lam = 10 ** (h2['CRVAL1'] + np.arange(h2['NAXIS1']) * h2['CDELT1']) * 10

        templates = np.empty((ssp.size, n_ages, n_metal))
        age_grid = np.empty((n_ages, n_metal))
        metal_grid = np.empty((n_ages, n_metal))
        flux = np.ones((n_ages, n_metal))

        for j, age in enumerate(ages):
            for k, met in enumerate(metals):
                p = all.index((age, met))
                hdu = fits.open(files[p])
                ssp = hdu[0].data
                templates[:, j, k] = ssp
                age_grid[j, k] = age
                metal_grid[j, k] = met

        new_wave = np.arange(lam[0], lam[-1], np.min(np.diff(lam)))
        new_templates = np.zeros([len(new_wave), templates.shape[1], templates.shape[2]])
        for i in range(templates.shape[1]):
            for j in range(templates.shape[2]):
                interpolator = interpolate.interp1d(lam, templates[:, i, j])
                new_templates[:, i, j] = interpolator(new_wave)

        new_templates = new_templates * flux_factor

        lam = new_wave
        templates = new_templates
        new_wave = None
        new_templates = None

        if age_range is not None:
            w = (age_range[0] <= age_grid[:, 0]) & (age_grid[:, 0] <= age_range[1])
            templates = templates[:, w, :]
            age_grid = age_grid[w, :]
            metal_grid = metal_grid[w, :]
            flux = flux[w, :]
            n_ages, n_metal = age_grid.shape

        if metal_range is not None:
            w = (metal_range[0] <= metal_grid[0, :]) & (metal_grid[0, :] <= metal_range[1])
            templates = templates[:, :, w]
            age_grid = age_grid[:, w]
            metal_grid = metal_grid[:, w]
            flux = flux[:, w]
            n_ages, n_metal = age_grid.shape

        self.templates = templates
        self.lam_temp = lam
        self.age_grid = age_grid
        self.metal_grid = metal_grid
        self.n_ages = n_ages
        self.n_metal = n_metal
        self.flux = flux
        self.FWHM_tem = FWHM_tem



import logging
import numpy as np
from astropy.io import fits


def write_cube(data_cube, params, x_edges, y_edges, new_wave, filepath, cube_idx, velscale, version):
    '''
    Write the final data cube values into a fits file, with proper header information written.
    :param data_cube:
    :param params:
    :param x_edges:
    :param y_edges:
    :param new_wave:
    :param filepath:Z
    :param cube_idx:
    :param velscale:
    :param version:
    :return:
    '''

    wave_range = params['ssp_params']['wave_range']
    # Do the wavelength cut
    if wave_range!=None:
        good_lam = (new_wave >= wave_range[0]) & (new_wave <= wave_range[1])
        new_wave = new_wave[good_lam]
        data_cube = data_cube[:, :, good_lam]

    # add noise on it
    if params['cube_params']['add_noise'] == True:
        logging.info('Adding noise on the flux with sn = %s' % params['cube_params']['sn'])
        noise = data_cube / params['cube_params']['sn']
        data_cube_n = np.random.normal(data_cube, noise)
    else:
        logging.info('No noise was added on the flux')
        data_cube_n = data_cube


    logging.info('Writing Headers...')
    # Write header
    hdr = fits.Header()
    hdr['MODEL'] = params['other_params']['model_name']
    hdr['DIST_kpc'] = params['oparams']['distance']
    hdr['l_deg'] = params['oparams']['l']
    hdr['b_deg'] = params['oparams']['b']
    hdr['theta_zx'] = params['oparams']['theta_zx']
    hdr['theta_yz'] = params['oparams']['theta_yz']
    hdr['XCOORD'] = params['cube_params']['x_coord']
    hdr['YCOORD'] = params['cube_params']['y_coord']
    hdr['SNR'] = params['cube_params']['sn']
    hdr['X_RES'] = params['cube_params']['spatial_resolution'][0]
    hdr['Y_RES'] = params['cube_params']['spatial_resolution'][1]
    hdr['NCPU'] = params['other_params']['ncpu']
    hdr['USELOSVD'] = params['cube_params']['use_losvd']
    hdr['USEEXTIN'] = params['cube_params']['use_extinc']
    hdr['DUSTFACT'] = params['cube_params']['extinc_factor']
    hdr['ADDNOISE'] = params['cube_params']['add_noise']
    hdr['BOOTSTRA'] = params['cube_params']['bootstrap_table']
    hdr['USEDIST'] = params['cube_params']['use_dist']
    hdr['VELSCALE'] = velscale
    hdr['VERSION'] = version

    hdr['SSP'] = params['ssp_params']['model']
    hdr['ISOCHRON'] = params['ssp_params']['isochrone']
    hdr['IMF'] = params['ssp_params']['imf']
    hdr['SLOPE'] = params['ssp_params']['slope']
    hdr['NOALPHA'] = params['ssp_params']['single_alpha']
    hdr['FWHMGAL'] = params['ssp_params']['FWHM_gal']
    hdr['INTERP'] = params['ssp_params']['spec_interpolator']

    hdr1 = fits.Header()
    hdr1['EXTNAME'] = 'DATA'
    hdr1['OBJECT'] = 'MWcube'
    hdr1['BUNIT'] = '10**(-20)*erg/s/cm**2/Angstrom'
    # hdr1['LOSVD'] = params['cube_params']['use_losvd']
    # hdr1['RADESYS'] = 'ICRS'
    hdr1['CRPIX1'] = (x_edges.shape[0] - 1) / 2
    hdr1['CRPIX2'] = (y_edges.shape[0] - 1) / 2
    hdr1['CD1_1'] = -np.diff(x_edges)[0]
    hdr1['CD1_2'] = 0
    hdr1['CD2_1'] = 0
    hdr1['CD2_2'] = np.diff(y_edges)[0]
    hdr1['CUNIT1'] = 'deg     '
    hdr1['CUNIT2'] = 'deg     '
    if params['cube_params']['x_coord'] == 'ra' and params['cube_params']['y_coord'] == 'dec':
        hdr1['CTYPE1'] = 'RA---TAN'
        hdr1['CTYPE2'] = 'DEC--TAN'
    elif params['cube_params']['x_coord'] == 'l' and params['cube_params']['y_coord'] == 'b':
        hdr1['CTYPE1'] = 'GLON-TAN'
        hdr1['CTYPE2'] = 'GLAT-TAN'
    hdr1['CRVAL1'] = np.mean(x_edges)
    hdr1['CRVAL2'] = np.mean(y_edges)
    hdr1['CTYPE3'] = 'AWAV'
    hdr1['CUNIT3'] = 'Angstrom'
    hdr1['CD3_3'] = new_wave[1] - new_wave[0]
    hdr1['CRPIX3'] = 1
    hdr1['CRVAL3'] = new_wave[0]
    hdr1['CD1_3'] = 0
    hdr1['CD2_3'] = 0
    hdr1['CD3_1'] = 0
    hdr1['CD3_2'] = 0
    hdr1['ADDNOISE'] = params['cube_params']['add_noise']


    empty_primary = fits.PrimaryHDU(header=hdr)
    image_hdu1 = fits.ImageHDU(data_cube_n.T, header=hdr1)
    image_hdu3 = fits.ImageHDU(new_wave, name='WAVE')

    # Write file
    logging.info('Combining hduls...')
    if params['cube_params']['add_noise'] == True:
        hdr2 = hdr1.copy()
        hdr2['EXTNAME'] = 'ERROR'
        image_hdu2 = fits.ImageHDU(noise.T, header=hdr2)
        hdul = fits.HDUList([empty_primary, image_hdu1, image_hdu2, image_hdu3])
    else:
        hdul = fits.HDUList([empty_primary, image_hdu1, image_hdu3])
    hdul.writeto(filepath + 'data_cube_' + str(cube_idx) + '.fits', overwrite=True)



import glob
import numpy as np
from astropy.io import fits



def age_metal(filename, age_grid, metal_grid):
    '''
    extract the age and metallicity from the name of file
    modified from miles_util.py of PPXF
    '''
    s = filename[:-4].split('_')
    age = age_grid[int(s[-2])]
    metal = metal_grid[int(s[-1])]
    return age, metal


class pegase():
    '''
    This is a class which is written for uploading the pegase-hr SSP models.
    The SSP models have been interpolated in a better grid, following the steps
    in Kacharov et al. 2018 using ulyss.
    The new age and metallicity grid is similar as the MILES SSP models.
    This file is modified from miles_util.py of PPXF
    '''
    def __init__(self, pathname, FWHM_tem=0.55,
                 age_range=None, metal_range=None):

        files = glob.glob(pathname + 'PHR*.fits')
        model = fits.open(files[0])
        ages = np.exp(model[1].data) / 1000
        metals = model[2].data
        n_ages, n_metal = len(ages), len(metals)


        # assert set(all) == set([(a, b) for a in ages for b in metals]), \
            # 'Ages and Metals do not form a Cartesian grid'

        # Extract the wavelength range and logarithmically rebin one spectrum
        # to the same velocity scale of the galaxy spectrum, to determine the
        # size needed for the array which will contain the template spectra.

        h = model[0].header
        lam = h['CRVAL1'] + np.arange(h['NAXIS1']) * h['CDELT1']
        ssps = model[0].data
        # lam_range_temp = lam[[0, -1]]
        # ssp_new, ln_lam_temp = util.log_rebin(lam_range_temp, ssp, velscale=velscale)[:2]
        # ssp_new = ssp

        # if norm_range is not None:
        #     norm_range = np.log(norm_range)
        #     band = (norm_range[0] <= lam) & (lam <= norm_range[1])

        templates = np.empty((ssps.shape[-1], n_ages, n_metal))
        age_grid = np.empty((n_ages, n_metal))
        metal_grid = np.empty((n_ages, n_metal))
        flux = np.ones((n_ages, n_metal))

        # Convolve the whole Vazdekis library of spectral templates
        # with the quadratic difference between the galaxy and the
        # Vazdekis instrumental resolution. Logarithmically rebin
        # and store each template as a column in the array TEMPLATES.

        # Quadratic sigma difference in pixels PEGASE-HR --> galaxy
        # The formula below is rigorously valid if the shapes of the
        # instrumental spectral profiles are well approximated by Gaussians.
        # if FWHM_gal is not None:
        #     FWHM_dif = np.sqrt(FWHM_gal**2 - FWHM_tem**2)
        #     sigma = FWHM_dif/2.355/0.2   # Sigma difference in pixels

        # Here we make sure the spectra are sorted in both [M/H] and Age
        # along the two axes of the rectangular grid of templates.
        for j, age in enumerate(ages):
            for k, met in enumerate(metals):
                ssp = ssps[k, j, :]
                ssp = ssp * 1e-7  # Convert the SSP unit to MILES unit to make them the same
                # if FWHM_gal is not None:
                #     if np.isscalar(FWHM_gal):
                #         if sigma > 0.1:   # Skip convolution for nearly zero sigma
                #             ssp = ndimage.gaussian_filter1d(ssp, sigma)
                #     else:
                #         ssp = util.gaussian_filter1d(ssp, sigma)  # convolution with variable sigma
                # ssp_new = util.log_rebin(lam_range_temp, ssp, velscale=velscale)[0]
                # ssp_new = ssp
                # if norm_range is not None:
                #     flux[j, k] = np.mean(ssp_new[band])
                #     ssp_new /= flux[j, k]
                templates[:, j, k] = ssp
                age_grid[j, k] = age
                metal_grid[j, k] = met

        if age_range is not None:
            w = (age_range[0] <= age_grid[:, 0]) & (age_grid[:, 0] <= age_range[1])
            templates = templates[:, w, :]
            age_grid = age_grid[w, :]
            metal_grid = metal_grid[w, :]
            n_ages, n_metal = age_grid.shape

        if metal_range is not None:
            w = (metal_range[0] <= metal_grid[0, :]) & (metal_grid[0, :] <= metal_range[1])
            templates = templates[:, :, w]
            age_grid = age_grid[:, w]
            metal_grid = metal_grid[:, w]
            n_ages, n_metal = age_grid.shape

        # if norm_range is None:
        #     templates /= np.median(templates)  # Normalize by a scalar

        self.templates = templates
        self.lam_temp = lam
        self.age_grid = age_grid
        self.metal_grid = metal_grid
        self.n_ages = n_ages
        self.n_metal = n_metal
        self.flux = flux
        self.FWHM_tem = FWHM_tem



import glob, re
import numpy as np
from astropy.io import fits




def age_metal(filename):
    """
    Extract the age and metallicity from the name of a file of
    the MILES library of Single Stellar Population models as
    downloaded from http://miles.iac.es/ as of 2022

    This function relies on the MILES file containing a substring of the
    precise form like Zm0.40T00.0794, specifying the metallicity and age.

    :param filename: string possibly including full path
        (e.g. 'miles_library/Eun1.30Zm0.40T00.0794.fits')
    :return: age (Gyr), [M/H]

    """
    # s = re.findall(r'Z[m|p][0-9]\.[0-9]{2}T[0-9]{2}\.[0-9]{4}', filename)[0]
    s = re.findall(r'Z[m|p]\d+\.\d+T\d+\.\d+', filename)[0]
    split_result = s.split("T")
    metal = split_result[0]
    age = float(split_result[1])
    if "Zm" in metal:
        metal = -float(metal[2:])
    elif "Zp" in metal:
        metal = float(metal[2:])

    return age, metal




class miles:


    def __init__(self, pathname, FWHM_tem=2.51, age_range=None,
                 metal_range=None):

        files = glob.glob(pathname)
        assert len(files) > 0, "Files not found %s" % pathname

        all = [age_metal(f) for f in files]
        all_ages, all_metals = np.array(all).T
        ages, metals = np.unique(all_ages), np.unique(all_metals)
        n_ages, n_metal = len(ages), len(metals)

        assert set(all) == set([(a, b) for a in ages for b in metals]), \
            'Ages and Metals do not form a Cartesian grid'

        # Extract the wavelength range and logarithmically rebin one spectrum
        # to the same velocity scale of the galaxy spectrum, to determine the
        # size needed for the array which will contain the template spectra.
        hdu = fits.open(files[0])
        ssp = hdu[0].data
        h2 = hdu[0].header
        lam = h2['CRVAL1'] + np.arange(h2['NAXIS1']) * h2['CDELT1']
        # lam_range_temp = lam[[0, -1]]
        # ssp_new, ln_lam_temp = util.log_rebin(lam_range_temp, ssp, velscale=velscale)[:2]
        # ssp_new = ssp

        # if norm_range is not None:
        #     norm_range = np.log(norm_range)
        #     band = (norm_range[0] <= lam) & (lam <= norm_range[1])

        templates = np.empty((ssp.size, n_ages, n_metal))
        age_grid, metal_grid, flux = np.empty((3, n_ages, n_metal))

        # Convolve the whole Vazdekis library of spectral templates
        # with the quadratic difference between the galaxy and the
        # Vazdekis instrumental resolution. Logarithmically rebin
        # and store each template as a column in the array TEMPLATES.

        # Quadratic sigma difference in pixels Vazdekis --> galaxy
        # The formula below is rigorously valid if the shapes of the
        # instrumental spectral profiles are well approximated by Gaussians.
        # if FWHM_gal is not None:
        #     FWHM_dif = np.sqrt(FWHM_gal ** 2 - FWHM_tem ** 2)
        #     sigma = FWHM_dif / 2.355 / h2['CDELT1']  # Sigma difference in pixels

        # Here we make sure the spectra are sorted in both [M/H] and Age
        # along the two axes of the rectangular grid of templates.
        for j, age in enumerate(ages):
            for k, met in enumerate(metals):
                p = all.index((age, met))
                hdu = fits.open(files[p])
                ssp = hdu[0].data
                # if FWHM_gal is not None:
                #     if np.isscalar(FWHM_gal):
                #         if sigma > 0.1:  # Skip convolution for nearly zero sigma
                #             ssp = ndimage.gaussian_filter1d(ssp, sigma)
                #     else:
                #         ssp = util.gaussian_filter1d(ssp, sigma)  # convolution with variable sigma
                # ssp_new = util.log_rebin(lam_range_temp, ssp, velscale=velscale)[0]
                # ssp_new = ssp
                # if norm_range is not None:
                #     flux[j, k] = np.mean(ssp_new[band])
                #     ssp_new /= flux[j, k]  # Normalize every spectrum
                templates[:, j, k] = ssp
                age_grid[j, k] = age
                metal_grid[j, k] = met

        if age_range is not None:
            w = (age_range[0] <= age_grid[:, 0]) & (age_grid[:, 0] <= age_range[1])
            templates = templates[:, w, :]
            age_grid = age_grid[w, :]
            metal_grid = metal_grid[w, :]
            flux = flux[w, :]
            n_ages, n_metal = age_grid.shape

        if metal_range is not None:
            w = (metal_range[0] <= metal_grid[0, :]) & (metal_grid[0, :] <= metal_range[1])
            templates = templates[:, :, w]
            age_grid = age_grid[:, w]
            metal_grid = metal_grid[:, w]
            flux = flux[:, w]
            n_ages, n_metal = age_grid.shape

        # The code below is the cause of the flux scale issue!!!!!
        # if norm_range is None:
        #     flux = np.median(templates[templates > 0])
        #     templates /= flux  # Normalize by a scalar


        self.templates = templates
        self.lam_temp = lam
        if 'loginterp' in pathname:
            self.age_grid = 10 ** (age_grid - 9)
        else:
            self.age_grid = age_grid
        self.metal_grid = metal_grid
        self.n_ages = n_ages
        self.n_metal = n_metal
        self.flux = flux
        self.FWHM_tem = FWHM_tem
